/*
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

static int ev=-1;
static int ui=-1;

static int uinput_create(void)
{
	int fd;
	int i;
	struct uinput_user_dev dev;

	if((fd=open("/dev/uinput",O_WRONLY|O_NONBLOCK))==-1)
	{
		perror("open /dev/uinput");
		goto err1;
	}

	if(ioctl(fd,UI_SET_EVBIT,EV_KEY))
	{
		perror("ioctl UI_SET_EVBIT EV_KEY");
		goto err2;
	}

	if(ioctl(fd,UI_SET_EVBIT,EV_SYN))
	{
		perror("ioctl UI_SET_EVBIT EV_SYN");
		goto err2;
	}

	for(i=1;i<=248;i++)if(ioctl(fd,UI_SET_KEYBIT,i))
	{
		perror("ioctl UI_SET_KEYBIT");
		goto err2;
	}

	memset(&dev,0,sizeof(dev));
	snprintf(dev.name,UINPUT_MAX_NAME_SIZE,"amt-keyboard");
	dev.id.bustype= BUS_USB;
	dev.id.vendor=0xdead;
	dev.id.product=0xf00d;
	dev.id.version=1;

	if(write(fd,&dev,sizeof(dev))!=sizeof(dev))
	{
		perror("write uinput dev");
		goto err2;
	}

	if(ioctl(fd,UI_DEV_CREATE))
	{
		perror("ioctl UI_DEV_CREATE");
		goto err2;
	}

	return fd;

err2:	close(fd);
err1:	return -1;
}

static void uinput_remove(int fd)
{
	if(ioctl(fd,UI_DEV_DESTROY))perror("ioctl UI_DEV_DESTROY");
	if(close(fd))perror("close uinput");
}

static int event_create(char *fn)
{
	int fd;

	if((fd=open(fn,O_RDONLY))==-1)
	{
		perror("open device");
		return -1;
	}

	if(ioctl(fd,EVIOCGRAB,(void*)1))
	{
		perror("ioctl EVIOCGRAB");
		close(fd);
		return -1;
	}

	return fd;
}

static void event_remove(int fd)
{
	if(ioctl(fd,EVIOCGRAB,(void*)0))perror("ioctl EVIOCGRAB");
	if(close(fd))perror("close");
}

static void usage(void)
{
	fprintf(stderr,"Usage: amtmapper [-f] [-d <event-device>}\n");
	exit(1);
}

static void doterm(int unused)
{
	uinput_remove(ui);
	event_remove(ev);
	exit(1);
}

int main(int argc,char *argv[])
{
	int c;
	int fg=0;
	char *fn=NULL;
	struct input_event e;

	while((c=getopt(argc,argv,"fd:"))!=-1)switch(c)
	{
	case 'f':
		fg=1;
		break;
	case 'd':
		fn=optarg;
		break;
	default:
		usage();
	}

	if(!fn)usage();

	if(!fg)if(daemon(0,0))
	{
		perror("daemon");
		return 1;
	}

	if((ev=event_create(fn))==-1)return 1;
	if((ui=uinput_create())==-1)
	{
		event_remove(ev);
		return 1;
	}

	signal(SIGINT,doterm);
	signal(SIGHUP,doterm);
	signal(SIGTERM,doterm);
	signal(SIGQUIT,doterm);

	while(1)
	{
		if(read(ev,&e,sizeof(e))!=sizeof(e))
		{
			perror("read event");
			break;
		}

		switch(e.type)
		{
		case EV_KEY:
			if(e.code==KEY_RO)e.code=KEY_102ND;
		case EV_SYN:
			if(write(ui,&e,sizeof(e))!=sizeof(e))
			{
				perror("write uinput");
				break;
			}
			break;
		}
	}

	signal(SIGINT,SIG_IGN);
	signal(SIGHUP,SIG_IGN);
	signal(SIGTERM,SIG_IGN);
	signal(SIGQUIT,SIG_IGN);

	uinput_remove(ui);
	event_remove(ev);

	return 0;
}
